;; Generated package description from datetime.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "datetime" "0.10.30.20250203.204701" "Parsing, formatting and matching timestamps" '((emacs "25.1") (extmap "1.1.1")) :commit "2601120d4d2857cdbad1cf8d4b84d77308920835" :authors '(("Paul Pogonyshev" . "pogonyshev@gmail.com")) :maintainer '("Paul Pogonyshev" . "pogonyshev@gmail.com") :keywords '("lisp" "i18n") :url "https://github.com/doublep/datetime")
